/*
 *  Author: tisyang.K
 *  Email : tisyang@gmail.com
 */
#include <iostream>
using namespace std;

template <typename T>
class Matrix
{
public:
	// This friend declaration make sure the different Matrix classes
	// can access each other's data sucessfully.
	template<typename Y>
	friend class Matrix;

    Matrix();
    ~Matrix();

    Matrix(T *array, size_t row, size_t columnh);

    Matrix(const Matrix<T> &another);
    Matrix<T> & operator=(const Matrix<T> &another);

	template <typename Y>
	Matrix(const Matrix<Y> &another);
	template <typename Y>
	Matrix<T> & operator=(const Matrix<Y> &another);

	// If the two matrixes can not be multipled(the former's column
	// doesn't equal the later's row), this operator returns an empty matrix.
	// Use template to make it possible like  
	// Matrix<double> multiply Matrix<int>.
	// If Matrix<int> Multiply Matrix<double> is needed,
	// You can first make a value-copy of Matrix<int> to be a Matrix<double>
	// by using template copy constructor, and then use this multiply operator.
	template <typename Y>
    Matrix<T>  operator*(const Matrix<Y> &another) const;

public:
    Matrix<double> Inverse();
    Matrix<T> Transpose();
	
	size_t size() const
	{
		return _row*_column;
	}

    size_t getRow() const
    {
        return _row;
    }

    size_t getColumn() const
    {
        return _column;
    }

    T getElement(size_t row, size_t column) const
    {
        return _array[(row - 1) * _column + column - 1];
    }

    void setElement(const T element, size_t row, size_t column)
    {
        _array[(row - 1) * _column + column - 1] = element;
    }   

public:
    void printElements(ostream & os = cout) const;

private:
    template <typename Y>
    void init(Y *array, size_t row, size_t column);

private:
    size_t _row;
    size_t _column;
    T *_array;
};

template <typename T>
Matrix<T>::Matrix()
    : _row(0), _column(0), _array(NULL) {}

template <typename T>
Matrix<T>::Matrix(T *array, size_t row, size_t column)
{
    init(array, row, column);
}

template <typename T>
template <typename Y>
void Matrix<T>::init(Y *array, size_t row, size_t column)
{
	if(row == 0||column == 0)
	{
		_row = _column = 0;
		_array = NULL;
		return;
	}
	_row = row;
    _column = column;
    _array = new T[row*column];
    for(int i = 0; i< row*column; i++)
        _array[i] = array[i];
}

template <typename T>
Matrix<T>::Matrix(const Matrix<T> &another)
{
    init(another._array,another._row, another._column);
}

template <typename T>
Matrix<T> & Matrix<T>::operator=(const Matrix<T> &another)
{
    if(this == &another)
        return *this;

    delete []_array;
    init(another._array,another._row, another._column);
    return *this;
}

template <typename T>
template <typename Y>
Matrix<T>::Matrix(const Matrix<Y> &another)
{
    init(another._array,another._row, another._column);
}

template <typename T>
template <typename Y>
Matrix<T> & Matrix<T>::operator=(const Matrix<Y> &another)
{
    if(this == &another)
        return *this;

    delete []_array;
    init(another._array,another._row, another._column);
    return *this;
}

// the return type is same to the former matrix's type, so as in the 
// comment for this function's declaration, if Matrix<int> Multiply Matrix<double> 
//is needed, You can first make a value-copy of Matrix<int> to be a Matrix<double>
// by using template copy constructor, and then use this multiply operator.
template <typename T>
template <typename Y>
Matrix<T>  Matrix<T>::operator*(const Matrix<Y> &another) const
{
    Matrix<T> matrix;
    if(this->_column != another._row)
        return matrix;

    size_t row = matrix._row = this->_row;
    size_t column = matrix._column = another._column;
    T * array = matrix._array = new T[row*column];

    for(int i=0; i<row; i++)
        for(int j=0; j<column; j++)
        {
            array[i*column+j] = 0;
            for(int k=0; k<another._row; k++)
                array[i*column+j] += this->_array[i*another._row + k]*another._array[j+ k*column];
        }

    return matrix;
}

template <typename T>
Matrix<T>::~Matrix()
{
    delete []_array;
}

template <typename T>
Matrix<double> Matrix<T>::Inverse()
{
    Matrix<double> empty;
    if(_row != _column)
        return empty;

    int n = _row;
    Matrix<double> matrix(*this);
    double *array = matrix._array;

    for(int k=0; k<n; k++)
    {
        for(int i=0; i<n; i++)
        {
            if(i != k)
            {
                if(array[k*n+k] == 0)
                    return empty;
                array[i*n+k] = -array[i*n+k]/array[k*n+k];
            }
        }
        if(array[k*n+k] == 0)
            return empty;
        array[k*n+k] = 1/array[k*n+k];
        for(int i=0; i<n; i++)
        {
            if(i != k)
            {
                for(int j=0; j<n; j++)
                {
                    if(j != k)
                        array[i*n+j] += array[k*n+j] * array[i*n+k];
                }
            }
        }
        for(int j=0; j<n; j++)
        {
            if(j != k)
                array[k*n+j] *= array[k*n+k];
        }
    }

    return matrix;
}

template <typename T>
Matrix<T> Matrix<T>::Transpose()
{
    if(_row == 0 || _column == 0)
        return *this;

    Matrix<T> copy(*this);
    T *MResult = copy._array;
    T *MSource = this->_array;
    int m = copy._row = _column;
    int n = copy._column =  _row;

    for(int i=0; i<m; i++)
        for(int j=0; j<n; j++)
            MResult[i*n+j] = MSource[j*m+i];

    return copy;
}

template <typename T>
void Matrix<T>::printElements(ostream & os) const
{
    os.clear();
    if(_row == 0 || _column == 0)
    {
        os<<"An empty matrix can not be printed!"<<endl;
        return;
    }
    for(int i=1; i<=_row; i++)
    {
        for(int j=1; j<=_column; j++)
            os<<'\t'<<getElement(i,j);
        os<<endl;
    }
}